# __init__.py
#
# Copyright 2015 Socos LLC
#

import cloud_storage, util


# Imports / Setup ##############################################################

# Core
import datetime
import dateutil.parser
import logging
import os
from pprint import pformat

from google.appengine.ext import ndb


LOG = logging.getLogger(__name__)


# NDB Base Model ###############################################################

class BaseModel(ndb.Model):
    """ Inherit from this model to support datastore_to_gcs transfers """

    last_modified = ndb.DateTimeProperty(required=True, auto_now=True)

    def serializable(self):
        obj = self.to_dict()
        obj = util.serializable(obj)
        if self.key:
            obj['id'] = self.key.id()
        return obj


# Interface ####################################################################

def dump(model, gcs_bucket, gcs_path, since=None):
    """ Creates a GCS file containing entites from a Datastore model

    Entities are individually JSON encoded and written one per line

    Parameters
    ----------
    model : ndb.Model
    gcs_bucket : str
        The name of the target GCS bucket
    gcs_path : str
        The name of the file to create
    since : datetime
        If provided, compared with model.last_modified to only get newly
        modified entities
    """
    if since:
        print 'since: {}'.format(since)
        new_items_iter = model.query(model.last_modified >= since).iter()
    else:
        new_items_iter = model.query().iter()
    new_items = [i.serializable() for i in new_items_iter]
    if len(new_items) > 0:
        LOG.info("Uploading {n} items in {m} to {o}".format(n=len(new_items),
                                                            m=model._get_kind(),
                                                            o=gcs_path))
        cloud_storage.upload_data(new_items, gcs_bucket, gcs_path)
    else:
        LOG.info("No items in {m} to transfer to {o}".format(m=model._get_kind(),
                                                             o=gcs_path))


def dump_log(model, gcs_bucket, gcs_log_dir, dump_all=False):
    """ Creates a timestamped file in GCS containing new entities

    Entities are individually JSON encoded and written one per line

    The timestamp of the previous log file is used to determine what date
    to pull entities from.

    Parameters
    ----------
    model : ndb.Model
    gcs_bucket : str
        The name of the target GCS bucket
    gcs_log_dir : str
        The name of the directory in the bucket to dump files to
    dump_all : bool
        If True, all entities of model are dumped, not just new ones
    """
    if dump_all:
        last_log = None
    else:
        last_log = _last_log_datetime(gcs_bucket, gcs_log_dir)
        print 'last log at {}'.format(last_log)
    log_name = cloud_storage.log_name(datetime.datetime.utcnow())
    gcs_object = os.path.join(gcs_log_dir, log_name)
    dump(model, gcs_bucket, gcs_object, since=last_log)
    return log_name


def update(model, gcs_bucket, gcs_object):
    """Updates the given GCS object with new data from the given model.

    Uses last_modified to determine the date to get items from. Bases the
    identity of entities in the GCS object on their 'id' field -- existing
    entities for which new data is found will be replaced.

    Parameters
    ----------
    model : ndb.Model
    gcs_bucket : str
    gcs_object : str
    """
    # If file doesn't exist, just dump
    if not cloud_storage.exists(gcs_bucket, gcs_object):
        LOG.info('No object to update, calling dump(...)')
        return dump(model, gcs_bucket, gcs_object)

    # Get preexisting items
    transferred_items = cloud_storage.download_object(gcs_bucket, gcs_object)
    LOG.info('{} items exist'.format(len(transferred_items)))

    # Find the most recently modified one
    last_date = datetime.datetime(1, 1, 1)
    for item in transferred_items:
        modified_date = dateutil.parser.parse(item['last_modified'])
        if modified_date > last_date:
            last_date = modified_date

    # Get all items modified after that date
    LOG.info('Last date on record: {}'.format(last_date.isoformat()))
    new_items_iter = model.query(model.last_modified > last_date).iter()
    new_items = tuple(item.serializable() for item in new_items_iter)
    new_items_by_id = {i['id']: i for i in new_items}

    if new_items:  # Found new items -- update existing items GCS
        items_by_id = {i['id']: i for i in transferred_items}
        items_by_id.update(new_items_by_id)
        items = items_by_id.values()
        LOG.info("Updating {n} items in {m} to {o}".format(n=len(new_items),
                                                           m=model._get_kind(),
                                                           o=gcs_object))
        cloud_storage.upload_data(items, gcs_bucket, gcs_object)
    else:  # Nothing to update with.
        LOG.info("No new items in {m} to append to {o}".format(m=model._get_kind(),
                                                               o=gcs_object))


# Private Methods ##############################################################

def _last_log_datetime(gcs_bucket, gcs_dir):
    """ Finds the last date for files under /gcs_bucket/gcs_dir

    Files must be named as ISO formatted datetime strings with the .json extension.

    Returns
    -------
    datetime.datetime

    Raises
    ------
    ValueError
        If a file in the directory has a noncompliant name
    """
    items_in_dir = cloud_storage.list_objects(gcs_bucket, gcs_dir)
    LOG.debug('Directory contains: {}'.format(items_in_dir))
    if len(items_in_dir) == 0:
        return None  # since=None just defaults to get all entries from datastore.

    # item[:-5] removes .json extension
    isostrings = [item[:-5] for item in items_in_dir]
    try:
        dts = [dateutil.parser.parse(item) for item in isostrings]
        return max(dts) + datetime.timedelta(seconds=1)
    except ValueError, e:
        LOG.error(u"Couldn't parse date in {}".format(pformat(isostrings)))
        raise e




# cloud_storage.py
#
# Copyright 2015 Socos LLC
#
"""
This is Google Cloud Storage client for Google App Engine. It's a wrapper
around cloudstorage.
"""

# Imports ######################################################################

from __future__ import print_function
import collections
import json
import logging
from pprint import pformat

import cloudstorage

from datastore_to_gcs import util


LOG = logging.getLogger(__name__)


# Interface ####################################################################

def list_objects(path_base, rest_of_path=''):
    """
    Lists the objects at the specified path in GCS.

    Parameters
    ----------
    path_base : str
    rest_of_path : str

    Returns
    -------
    list[str]
        A list of filenames

    """
    path_to_dir = util.parse_cloud_storage_path(path_base, rest_of_path)
    path_to_dir = path_to_dir[:-1] if path_to_dir[-1] == '/' else path_to_dir
    items = cloudstorage.listbucket(path_to_dir)
    relative_paths = (i.filename[len(path_to_dir) + 1:] for i in items if not i.is_dir)
    # @bistenes: Can't tell you why, but the above line also produces an
    #   entry for the directory despite the is_dir check, but *only in
    #   production*. Couldn't repro in testing or staging. The relative
    #   path from the directory to itself, as computed by the line above,
    #   is of course ''. Hence the following return, which filters out
    #   those entries.
    return [r for r in relative_paths if r != '']


def exists(path_base, rest_of_path=''):
    objs = list_objects(path_base)
    return rest_of_path in objs


def download_object(path_base, rest_of_path, needed_fields=(), object_class=util.DataDict):
    """
    Loads the data at file_path into memory, interpreting it as lines of JSON

    Parameters
    ----------
    path_base : str
        The bucket or directory of the object to be downloaded
    rest_of_path : str
        The path of the object to be downloaded
    needed_fields : list[str] | tuple[str]
        The fields to filter each object down to. Nested fields are supported
        using '.' as separator, i.e. in `{'foo': {'bar': 'baz'}}`, the `bar`
        field can be filtered for by passing `needed_fields=['foo.bar']`
    object_class : type
        A class to wrap each line of data in

    Returns
    -------
    list[object_class]
        Representing the data in the downloaded object
    """
    gcs_file = cloudstorage.open(util.parse_cloud_storage_path(path_base, rest_of_path))
    item_dicts = _objects_from_file(gcs_file, needed_fields)
    gcs_file.close()
    return _parse_result(item_dicts, object_class)


def upload_data(input_list, path_base, rest_of_path):
    """
    Uploads input_list to a file at the given path.

    Each item of input_list gets passed through util.serializable
    and then serialized as JSON. Each item gets one line in the file, producing
    a file for which each line is a valid JSON object.

    Parameters
    ----------
    input_list : list[object]
        Non-empty iterable of objects to be serialized
    path_base : str
        The first part of the path, including the bucket name
    rest_of_path : str
        The rest of the path, including the filename

    Returns
    -------
    The (bucket, object) path of the uploaded object, or None if failed.
    """
    assert isinstance(input_list, collections.Iterable)
    file_path = util.parse_cloud_storage_path(path_base, rest_of_path)
    # Abort if input_list is empty
    if not any(True for _ in input_list):
        LOG.warn('Refusing to upload empty file to ' + file_path)
        return None
    # Carry out the upload
    gcs_file = cloudstorage.open(file_path, mode='w', content_type='text/plain')
    for item in input_list:
        ser_item = util.serializable(item)
        json.dump(ser_item, gcs_file, encoding='utf-8')
        gcs_file.write(_encode('\n'))
    gcs_file.close()
    return util.parse_cloud_storage_path_split(path_base, rest_of_path)


def log_name(utc_dt):
    dt_string = utc_dt.replace(microsecond=0).isoformat()
    return dt_string + '.json'


# Implementation ###############################################################

def _encode(raw):
    return raw.encode('utf-8')


def _decode(utf8):
    return utf8.decode('utf-8')


def _objects_from_file(file_object, needed_fields):
    file_object.seek(0)
    items = []
    for linenum, line in enumerate(file_object):
        try:
            items.append(json.loads(line))
        except ValueError as e:
            raise ValueError('Error at line %d: %s' % (linenum, e))
    if len(needed_fields) > 0:
        for i in range(len(items)):
            items[i] = util.filter_paths(items[i], [p.split('.') for p in needed_fields])
    return items


def _parse_result(item_dicts, object_class):
    result = []
    for item_dict in item_dicts:
        try:
            result.append(object_class(**item_dict))
        except TypeError, e:
            LOG.error('TypeError on\n{}'.format(pformat(item_dict, indent=4)))
            raise e
    return result




from datetime import date, datetime
import enum
from toolz.dicttoolz import get_in, assoc_in

from google.appengine.ext.ndb import Key


def filter_paths(item, paths):
    filtered_item = {}
    for key_path in paths:
        value = get_in(key_path, item)
        filtered_item = assoc_in(filtered_item, key_path, value)
    return filtered_item


def parse_cloud_storage_path(base, rest):
    if base[0] != '/':
        base = '/' + base
    if base[-1] == '/':
        base = base[:-1]
    if rest and rest[0] == '/':
        rest = rest[1:]
    return '/'.join((base, rest)) if rest else base


def parse_cloud_storage_path_split(base, rest):
    if base[0] == '/':
        base = base[1:]
    if base[-1] == '/':
        base = base[0:-1]
    base_parts = base.split('/')
    bucket = base_parts[0]
    object_name = ''
    if len(base_parts) > 1:
        object_name = '/'.join(base_parts[1:]) + '/'
    if rest[0] == '/':
        rest = rest[1:]
    object_name += rest
    return bucket, object_name


def serializable(item):
    """Converts item into an equivalent JSON serializable object.

    Parameters
    ----------
    item : T

    Returns
    -------
    dict
    """
    # Okay, go through the actual recursive serialization
    if isinstance(item, dict):
        result = {}
        for key, val in item.items():
            result[key] = serializable(val)
    elif isinstance(item, (list, tuple)) or hasattr(item, '__iter__'):
        result = []
        for i in item:
            result.append(serializable(i))
    elif isinstance(item, datetime) or isinstance(item, date):
        result = item.isoformat()
    elif isinstance(item, Key):
        result = item.id()
    elif isinstance(item, enum.Enum):
        result = item.name
    elif hasattr(item, 'serializable') and callable(item.serializable):
        result = item.serializable()
    elif hasattr(item, '__dict__'):
        # Note that this doesn't include class attributes
        result = serializable(item.__dict__)
    else:
        result = item

    return result


class DataDict(dict):
    """ A wrapper for dict which prints the object in case of a KeyError. """

    def __getitem__(self, key):
        try:
            val = dict.__getitem__(self, key)
        except KeyError:
            raise KeyError('{0} not found in {1}'.format(key, self.keys()))

        return val

    def __getattr__(self, key):
        try:
            val = dict.__getitem__(self, key)
            return val
        except KeyError:
            raise AttributeError('{0} not found in {1}'.format(key, self.keys()))


class CommonEqualityMixin(object):

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)




# setup.py
#
# Copyright 2017 Socos LLC
#

import os

from setuptools import setup

def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as f:
        return f.read()

setup(
    name='datastore-to-gcs',
    packages=['datastore_to_gcs'],
    install_requires=['enum', 'GoogleAppEngineCloudStorageClient', 'python-dateutil', 'toolz'],
    version='0.3',
    license='Apache License 2.0',
    description='Transfer data from Google Cloud Datastore to Google Cloud Storage.',
    long_description=(read('README.rst')),
    author='Brandon Istenes',
    author_email='bistenes@socos.me',
    url='https://github.com/SocosLLC/datastore-to-gcs',
    download_url='https://github.com/SocosLLC/datastore-to-gcs/archive/0.3.tar.gz',
    keywords=['gae', 'gcp', 'gcs', 'gcd', 'datastore', 'cloudstorage', 'transfer', 'dump'],
    classifiers=['Development Status :: 4 - Beta',
                 'License :: OSI Approved :: Apache Software License']
)
